/*======================================================================*
 *  Module .............btdiag
 *  File ...............main.c
 *  Author .............Traveler Hauptman
 *  Creation Date ......18 Mar 2003
 *                                                                      *
 *  ******************************************************************  *
 *                                                                      *
 *  NOTES:
 *
 *  REVISION HISTORY:
 *  18 Mar 2003 - TH
 *    File created    
 *  25 Mar 2003 - BZ
 *    File documented    
 *                                                                      *
 *======================================================================*/

/** \file main.c
    Provides a dignostic interface to a set of motor controllers.
    
 */


/*==============================*
 * INCLUDES - System Files      *
 *==============================*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <syslog.h>
#include <stddef.h>
#include <signal.h>
#include <pthread.h>
#include <errno.h>
#include <semaphore.h>
#include <malloc.h>
#include <process.h>
#include <time.h>
#include <inttypes.h>

/* QNX specific header files */
#include <sys/sched.h>
#include <sys/mman.h>
#include <sys/neutrino.h>
#include <sys/netmgr.h>
#include <sys/syspage.h>

/*==============================*
 * INCLUDES - Project Files     *
 *==============================*/
#include "globals.h"
#include "control_loop.h"
#include "btcan.h"
#include "btsystem.h"
#include "btwam.h"
#include "SimpleControl.h"
#include "playlist.h"
#include <curses.h>

/*==============================*
 * PRIVATE DEFINED constants    *
 *==============================*/
#define disp_offset 15
#define disp_width 11

/*==============================*
 * PRIVATE MACRO definitions    *
 *==============================*/

/*==============================*
 * PRIVATE typedefs and structs *
 *==============================*/
 
 /*==============================*
 * GLOBAL file-scope variables  *
 *==============================*/
char divider[disp_offset + (disp_width * MAX_NODES)];
SC_move_list ml;
control_thd_parms ctp;
SimpleCtl *s;
actuator_struct *act;
int num_act;
bus_struct *bus;
int num_bus;
uint64_t CPU_cycles_per_second;
int done = 0;
int motor_position[7];
pthread_mutex_t disp_mutex;
pthread_t display_thd_id;
int WACKYplay = 0;


//=======================================================

int npucks, cpuck;
int node;

int use_safety_module;
int entryLine;
int zeroing = 0, entry = 0;
int disp_puck = 0, disp_help = 0;
int useTRC = 0;
int saved_bus_error = 0;

int indexes[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int offsets[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int modes[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
long int temperatures[15] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
double commands[15];
double newpid[15];
double A,B,C,D;
enum {HELP_SCREEN, MAIN_SCREEN, GCOMP_SCREEN, HOMING_SCREEN, TIMING_SCREEN, PLAYLIST_SCREEN};
int present_screen = 1;
int startup_stage = 0; //Set to 1 when main loop is started. Used to control sigint handler
//==============================================================

/*==============================*
 * PRIVATE Function Prototypes  *
 *==============================*/
void RenderScreen();
void RenderHelp();
void RenderScreenGcomp();
void ProcessInput(int c);
void Initialization();
void Shutdown();
void DisplayThread();
void StartDisplayThread();
void WACKYeval();
char *ifstr(int in, char *yes, char *no);

/*==============================*
 * Functions                    *
 *==============================*/
 
/** Traps the Ctrl-C signal.
    Quits the program gracefully when Ctrl-C is hit.
    
*/
void sigint_handler()
{
  if (!startup_stage)
  {
    endwin();
    exit(1);
  }
  else
    done = 1;
}

/** Entry point for the application.
    Initializes the system and executes the main event loop.
    
*/
int main( int argc, char **argv ) //{{{
{

    char chr;
    char filename[50];
    int cnt, cnt2, num;
    int ret;
    int status[MAX_NODES];

    Initialization();

    ret = InitWAM("wam.dat");
    if (ret)
    {
      CloseSystem();
      exit(0);
    }
    
    DumpData2Syslog();
    
    act = GetActuators(&num_act); //get pointer and number of actuators
    npucks = num_act;
    

    bus = GetBuses(&num_bus);
    s = GetWAMsc();
    MLconstruct(&ml, s, 7, 50);
    cnt2 = 0;
    ret = 1;

    start_control_threads(26, 2, WAMControlThread, NULL);
    signal (SIGINT, sigint_handler);
    startup_stage = 1;
    StartDisplayThread();
    
    clear();
    refresh();
    while (!done)
    {
        WACKYeval();
        MLeval(&ml);
        if ((chr = getch()) != ERR)
            ProcessInput(chr);
        usleep(100000);
    }

    Shutdown();
    CloseSystem();


}

/** Initialize curses, syslog, and the display mutex.
    
*/
void Initialization() //{{{
{
    int err;
    int i,cnt;

    /* Get the cycles-per-second of this machine */
    CPU_cycles_per_second = SYSPAGE_ENTRY(qtime)->cycles_per_sec;

    /* Initialize the ncurses screen library */
    initscr();
    addstr("screen initalized\n");
    refresh();
    cbreak();
    noecho();
    timeout(0);
    clear();
    
    //* Initialize syslog */
    openlog("WAM", LOG_CONS | LOG_NDELAY, LOG_USER);
    addstr("syslog initalized\n");
    refresh();
    syslog(LOG_ERR, "syslog initalized by WAM application");
    
    /* Initialize the display mutex */
    err = pthread_mutex_init(&(disp_mutex), NULL);
    if (err)
    {
        syslog(LOG_ERR, "Could not initialize  mutex for displays.");
    }

    /* Set the present active puck to 0 */
    cpuck = 0;
    
    
    /* Create the divider line */
    divider[0] = '\0';
    for(cnt = 0; cnt < disp_offset + npucks * disp_width; cnt++)
    {
        strcat(divider, "-");
    }
    
} //}}}

/** Shut the application down gracefully.
    Stops the control threads, ncurses, syslog, and frees the movelist.
*/
void Shutdown() //{{{
{

    stop_control_threads();
    //curses
    

    //syslogd
    MLdestroy(&ml);
    closelog();
    endwin();
} //}}}

/** Starts the display thread.
    Creates a new thread, sets its priority, and runs it as a display thread.
*/
void StartDisplayThread()
{
    pthread_attr_t display_attr;
    struct sched_param display_param;

    pthread_attr_init(&display_attr);
    pthread_attr_setinheritsched(&display_attr, PTHREAD_EXPLICIT_SCHED);
    display_param.sched_priority = 10;
    pthread_attr_setschedparam(&display_attr, &display_param);

    pthread_create(&display_thd_id, &display_attr, (void*)DisplayThread, 0);
    sched_setscheduler(display_thd_id, SCHED_FIFO, &display_param);
    if (display_thd_id == -1)
    {
        syslog(LOG_ERR, "start_display_thread:Couldn't start display thread!");
        exit( -1);
    }
}

/** Spins in a loop, updating the screen.
    Runs as its own thread, updates the screen.
*/
void DisplayThread()
{
    int err;
    clear();
    refresh();
    while (!done)
    {
        if ((err = pthread_mutex_lock( &(disp_mutex) )) != EOK)
        {
            syslog(LOG_ERR, "display mutex failed: %d", err);
        }
        switch (present_screen)
        {
          case HELP_SCREEN:
            RenderHelp();
            break;
          case MAIN_SCREEN:
            RenderScreen();
            break;
          case GCOMP_SCREEN:
            RenderScreenGcomp();
          break;
            
        }

        pthread_mutex_unlock( &(disp_mutex) );
        usleep(100000);
    }

}

/** Locks the display mutex.
    Allows the user to enter on-screen data without fear of display corruption.
*/
void start_entry()
{
    int err;
    if ((err = pthread_mutex_lock( &(disp_mutex) )) != EOK)
    {
        syslog(LOG_ERR, "display mutex failed: %d", err);
    }
    move(entryLine, 1);
    echo();
    timeout( -1);
}

/** Unlocks the display mutex.
    Allows the computer to resume automatically updating the screen.
*/
void finish_entry()
{
    noecho();
    timeout(0);
    move(entryLine, 1);
    addstr("                                                                              ");
    refresh();
    pthread_mutex_unlock( &(disp_mutex) );
}

/** ???.
    
*/
char *ifstr(int in, char *yes, char *no)
{
    if (in)
        return yes;
    else
        return no;
}

/** Show the help screen.
    Displays the on-screen help.
*/
void RenderHelp() //{{{
{
    int val;
    int cnt, idx;
    uint64_t result;
    int line;

    
    line = 0;
    mvprintw(line , 0, "'a' Power wam");
    ++line;
    mvprintw(line , 0, "'d' Idle wam");
    ++line;
    mvprintw(line , 0, " ");
    ++line;
    mvprintw(line , 0, "'[' Select previous joint");
    ++line;
    mvprintw(line , 0, "']' Select next joint");
    ++line;

    refresh();
}

/** Draw the main information screen.
    Dynamically draw the puck information on the screen.
*/
void RenderScreen() //{{{
{
    int val;
    int cnt, idx;
    uint64_t result;
    int line;
    wam_vector Mpos, Mtrq, Jpos, Jtrq, GCtrq,GCpos, t2, t3, t4;
    wam_vector pos,trq;
    double pi=3.14159,a,b,c,d;

    line = 0;
    mvprintw(line , 0, "BarrettTech Diagnostics (btdiag)"); ++line;
    if (disp_puck)
    {
        mvprintw(line , 0, "       PUCK: "); ++line;
        mvprintw(line , 0, "        Pos: "); ++line;
        mvprintw(line , 0, "        Trq: "); ++line;
    }
    else
    {
        mvprintw(line , 0, "      JOINT: "); ++line;
        mvprintw(line , 0, "      Angle: "); ++line;
        mvprintw(line , 0, "     Torque: "); ++line;
    }
    mvprintw(line, 0, "%s", divider); ++line;
    mvprintw(line, 0, "[c] Command: "); ++line;
    mvprintw(line, 0, "%s", divider); ++line;
    mvprintw(line , 0, "[%s] PIDmode: ",ifstr(s[0].mode,"1","2")); ++line;
    mvprintw(line , 0, "    PID cmd: "); ++line;
    mvprintw(line , 0, "[p] Kp     : "); ++line;
    mvprintw(line , 0, "    Kd     : "); ++line;
    mvprintw(line , 0, "    Ki     : "); ++line;
    mvprintw(line , 0, "[v] Vel    : "); ++line;
    mvprintw(line , 0, "    Acc    : "); ++line;
    mvprintw(line , 0, "[G] Temp C : "); ++line;
    //mvprintw(line , 0, "[E] Vel    : "); ++line;
    mvprintw(line, 0, "%s", divider); ++line;
    mvprintw(line , 0, "[t] Trjmode: "); ++line;
    mvprintw(line , 0, "    Trjtarg: "); ++line;
    mvprintw(line, 0, "%s", divider); ++line;
    mvprintw(line, 0, "Play List    [/]Stop [.]Play [>]Next [<]Prev [+]Add [-]Del [F]Save [f]Load"); ++line;
    mvprintw(line, 0, "     Status: %8s Current pt: %2d Total pts: %3d", ifstr(MLrunning(&ml), "Running", "Stopped"),ml.Cmove+1,ml.Nmoves); ++line;
    mvprintw(line, 0, "       This: ");

    getWAMjoints(&Mpos, &Mtrq, &Jpos, &Jtrq);
    getWAMmotor_position(motor_position);
    calcGcomp2(&t2, &t3, &t4);

    fer (cnt, num_act)
    {
        line = 1;
        idx = act[cnt].puck.ID;
        if (disp_puck)
        {
            mvprintw(line , disp_offset + disp_width*cnt, "   %2d   ", idx); ++line;
            mvprintw(line , disp_offset + disp_width*cnt, "%8d ", act[cnt].puck.position); ++line;
            mvprintw(line , disp_offset + disp_width*cnt, "%8d ", act[cnt].puck.torque_cmd); ++line;
        }
        else
        {
            if (motor_position[cnt] == motor_position[cpuck])
            {
                mvprintw(line , disp_offset + disp_width*cnt, " -[%2d]- ", motor_position[cnt]+1); ++line;
            }
            else
            {
                mvprintw(line , disp_offset + disp_width*cnt, "   %2d  ", motor_position[cnt]+1); ++line;
            }
            mvprintw(line, disp_offset + disp_width*cnt, "%8.4f ", Jpos.q[motor_position[cnt]]); ++line;
            mvprintw(line, disp_offset + disp_width*cnt, "%8.4f ", Jtrq.q[motor_position[cnt]]); ++line;
        }
        ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", commands[cnt]); ++line;
        ++line;
        if (s[motor_position[cnt]].mode == 0)
        {
            mvprintw(line , disp_offset + disp_width*cnt, "OFF  "); ++line;
        }
        else if (s[motor_position[cnt]].mode == 2)
        {
            mvprintw(line , disp_offset + disp_width*cnt, "ON    "); ++line;
        }
        else
        {
            mvprintw(line , disp_offset + disp_width*cnt, "UNK:%d ", s[motor_position[cnt]].mode);      ++line;
        }

        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", s[motor_position[cnt]].pid.yref);    ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", s[motor_position[cnt]].pid.Kp );    ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", s[motor_position[cnt]].pid.Kd );    ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", s[motor_position[cnt]].pid.Ki);    ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", s[motor_position[cnt]].trj.vel);    ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", s[motor_position[cnt]].trj.acc);    ++line;
        mvprintw(line, disp_offset + disp_width*cnt, "%8.1f ", temperatures[act[cnt].puck.ID] / 10.0);    ++line;
        //mvprintw(line, disp_offset + disp_width*cnt, "%8ld ", vel[act[cnt].puck.ID]);    ++line;
        ++line;
        if (zeroing)
        {
            mvprintw(line, disp_offset + disp_width*cnt, "ZEROING");      ++line;
        }
        else if (s[motor_position[cnt]].trj.state == 0)
        {
            mvprintw(line, disp_offset + disp_width*cnt, "STOPPED");      ++line;
        }
        else if (s[motor_position[cnt]].trj.state == 1)
        {
            mvprintw(line, disp_offset + disp_width*cnt, "RUNNING");      ++line;
        }
        else
        {
            mvprintw(line, disp_offset + disp_width*cnt, "UNK:%d  ", s[motor_position[cnt]].trj.state);      ++line;
        }
        mvprintw(line, disp_offset + disp_width*cnt, "%8.3e ", s[motor_position[cnt]].trj.end);  ++line;
        ++line;
        ++line;
        ++line;
        if (ml.Nmoves)
            mvprintw(line, disp_offset + disp_width*cnt, "%8.4f ", ml.next[cnt].pos);
        else
            mvprintw(line, disp_offset + disp_width*cnt, " No Pts ");
        ++line;
        // mvprintw(line, disp_offset + disp_width*cnt, "%8.4f ", act[motor_position[cnt]].angle);
        //  ++line;
        //  mvprintw(line, disp_offset + disp_width*cnt, "%8.2e ", Jpos.q[motor_position[cnt]]);
        //   ++line;
    }
    if (useTRC)
      mvprintw(line, 0, " Torque ripple compensation is active! ");
      else
      mvprintw(line, 0, "                                         ");
    ++line;
    if (getGcomp())
      mvprintw(line, 0, " Gravity compensation is active! ");
      else
      mvprintw(line, 0, "                                         ");
    ++line;
    if (saved_bus_error)
       mvprintw(23,0,"Last Bus Error: %x\t\t", saved_bus_error);++line;

   
    entryLine = line;
    refresh();
} //}}}

void RenderScreenGcomp() //{{{
{
    int val;
    int cnt, idx;
    uint64_t result;
    int line;
    wam_vector Mpos, Mtrq, Jpos, Jtrq, GCtrq,GCpos, t2, t3, t4;
    wam_vector pos,trq;
    double pi=3.14159,a,b,c,d;

    line = 0;
    mvprintw(line, 0, "Communication Time %10.5f milliseconds  ", TurnaroundTime(0)); ++line;
    getGcompValues(&A,&B,&C,&D);
    mvprintw(line, 0, "Gcomp Calc-Coefficients: A: %8.4f B: %8.4f C: %8.4f D: %8.4f ", A,B,C,D); ++line;
    calcGcomp(&Jpos,&GCtrq,A,B,C,D);
    mvprintw(line, 0, "Gcomp Calc-Results     : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", GCtrq.q[0],GCtrq.q[1],GCtrq.q[2],GCtrq.q[3]); ++line;
    mvprintw(line, 0, "Gcomp Calc-Angles      : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", Jpos.q[0],Jpos.q[1],Jpos.q[2],Jpos.q[3]); ++line;
    set_wam_vector(&GCpos,0,-pi/2,-pi/2,0,0,0,0);
    a = -0.638225;
    b = -6.125;
    c = 0.700088;
    d = 9.8;
    calcGcomp(&GCpos,&GCtrq,a,b,c,d);
    mvprintw(line, 0, "Reference Gcomp Calc-Coefficients: A: %8.4f B: %8.4f C: %8.4f D: %8.4f ", a,b,c,d); ++line;
    mvprintw(line, 0, "Reference Gcomp Calc-Angles      : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", Jpos.q[0],Jpos.q[1],Jpos.q[2],Jpos.q[3]); ++line;
    mvprintw(line, 0, "Reference Gcomp Calc-Results     : 1: %8.4f 2: %8.4f 3: %8.4f 4: %8.4f ", GCtrq.q[0],GCtrq.q[1],GCtrq.q[2],GCtrq.q[3]); ++line;
   
    mvprintw(line, disp_offset, "               "); ++line;
    calcGcomp2(&t2, &t3, &t4);
    mvprintw(line, 0, "t2: %10.5f, %10.5f, %10.5f, %10.5f ", t2.q[0], t2.q[1], t2.q[2], t2.q[3]); ++line;
    mvprintw(line, 0, "t3: %10.5f, %10.5f, %10.5f, %10.5f ", t3.q[0], t3.q[1], t3.q[2], t3.q[3]); ++line;
    mvprintw(line, 0, "t4: %10.5f, %10.5f, %10.5f, %10.5f ", t4.q[0], t4.q[1], t4.q[2], t4.q[3]); ++line;
    ++line;
    mvprintw(line, 0, "Position offsets: %d, %d, %d, %d ", offsets[0], offsets[1], offsets[2], offsets[3]);
    entryLine = line;
    refresh();
}

/** Process user input.
    Handles the user's keypress, and performs the function desired by the user.
*/
void ProcessInput(int c) //{{{ Takes last keypress and performs appropriate action.
{
    int val;
    int tmp;
    int cnt;
    int ret;
    int idx, err;
    long ltmp;
    char chr;

    double ftmp;
    char fn[50];

    switch (c)
    {
	case '~':WACKYplay = !WACKYplay;
	    break;
        case 'x': /* eXit */
	case 'X':
            done = 1;
            break;
        case 'a': /* Activate WAM */
            PowerWAM();
	    setProperty(&(bus[act[0].bus].CANdev), SAFETY_MODULE, SAVED_BUS_ERR, 0);//clear the last bus error
            break;
        case 'd': /* iDle WAM */
            IdleWAM();
            break;
        case '[': /* Select previous puck */
            cpuck--;
            if (cpuck < 0)
                cpuck = npucks - 1;
            break;
        case ']': /* Select next puck */
            cpuck++;
            if (cpuck >= npucks)
                cpuck = 0;
            break;
        case 'Z': /* Zero WAM */
            zeroing = 1;
            ZeroWAM();
	    SetTRC(1);
            zeroing = 0;
            break;
        case 'A': /* Get Index */
            fer(cnt, npucks)
            {
                indexes[act[cnt].puck.ID] = -1;
                indexes[act[cnt].puck.ID] = GetProp(cnt, DRIVE_MODE);
            }
            break;
        case 'M': /* get Mode */
            fer(cnt, npucks)
            {
                modes[act[cnt].puck.ID] = -1;
                tmp = GetProp(cnt, CONTROL_MODE);
                //syslog(LOG_ERR,"%d Val1 %d",cnt,tmp);
                modes[act[cnt].puck.ID] = tmp;
                //syslog(LOG_ERR,"Val2 %d",modes[act[cnt].puck.ID]);
            }
            break;
        case '|': /*Get gravity compensation values */
            getGcompvalues();
            break;
        /*case '`':
            GCompCollectData("gcompdata.dat",2,2,1);
            break;*/
        case '\\': /* Toggle gravity compensation */
            toggleGcomp();
            break;
        case 'r': /* Toggle Torque Ripple Compensation */
            useTRC = !useTRC;
            SetTRC(useTRC);
            break;
        case 'W': /* Set mode to MODE_TORQUE for all pucks*/
            fer(cnt, npucks)
            SCsetmode(&(s[motor_position[cnt]]), MODE_TORQUE);
            break;
        case 'w': /* Set mode to MODE_TORQUE for present puck */
            SCsetmode(&(s[motor_position[cpuck]]), MODE_TORQUE);
            break;
        case ' ': /* Toggle puck/joint display */
            disp_puck = !disp_puck;
            break;
        case '_': /* Refresh display */
            if ((err = pthread_mutex_lock( &(disp_mutex) )) != EOK)
            {
                syslog(LOG_ERR, "display mutex failed: %d", err);
            }
            clear();
            pthread_mutex_unlock( &(disp_mutex) );
            break;
        case 'q': /* Set mode to MODE_IDLE for present puck */
            SCsetmode(&(s[motor_position[cpuck]]), MODE_IDLE);
            break;
        case 'Q': /* Set mode to MODE_IDLE for all pucks */
            fer(cnt, npucks)
            SCsetmode(&(s[motor_position[cnt]]), MODE_IDLE);
            break;
        case 'p': /* Enter PID constants for present puck \bug Check to see if we are in control mode 0 and able to change gains*/
            start_entry();
            addstr("Enter pid constants \"Kp Kd Ki\": ");
            refresh();
            ret = scanw("%lf %lf %lf %lf", &(newpid[0]), &(newpid[1]), &(newpid[2]), &(newpid[3]));
            //if (ret == 3)
            SCsetpid(&(s[motor_position[cpuck]]), newpid[0], newpid[1], newpid[2], newpid[3]);
            finish_entry();
            break;
        case 'P': /* Enter PID constants for all pucks \bug Check to see if we are in control mode 0 and able to change gains*/
	    start_entry();
            addstr("Enter pid constants \"Kp Kd Ki\": ");
            refresh();
            ret = scanw("%lf %lf %lf %lf", &(newpid[0]), &(newpid[1]), &(newpid[2]), &(newpid[3]));
            //if (ret == 3)
            fer(cnt, npucks)
            SCsetpid(&(s[motor_position[cnt]]), newpid[0], newpid[1], newpid[2], newpid[3]);
            finish_entry();
            break;
        case 'k': /* Enter Kill torque for present puck */
            start_entry();
            addstr("Enter Kill torque: ");
            refresh();
            ret = scanw("%lf", &(newpid[0]));
            s[motor_position[cpuck]].kill_torque = newpid[0];
            finish_entry();
            break;
        case 'K': /* Enter Kill torque for all pucks */
            start_entry();
            addstr("Enter Kill torque: ");
            refresh();
            ret = scanw("%lf", &(newpid[0]));
            fer(cnt, npucks)
            s[motor_position[cnt]].kill_torque = newpid[0];
            finish_entry();
            break;
        case 'v': /* Enter trajectory constants for present puck */
            start_entry();
            addstr("Enter trajectory constants \"Velocity Acceleration\": ");
            refresh();
            ret = scanw("%lf %lf", &(newpid[0]), &(newpid[1]));
            SCsettrjprof(&(s[motor_position[cpuck]]), newpid[0], newpid[1]);
            finish_entry();
            break;
        case 'V': /* Enter trajectory constants for all pucks */
            start_entry();
            addstr("Enter trajectory constants \"Velocity Acceleration\": ");
            refresh();
            ret = scanw("%lf %lf", &(newpid[0]), &(newpid[1]));
            fer(cnt, npucks)
            SCsettrjprof(&(s[motor_position[cnt]]), newpid[0], newpid[1]);
            finish_entry();
            break;

        case 't': /* Begin Trajectory control on present puck */
            SCstarttrj(&(s[motor_position[cpuck]]), commands[cpuck]);
            break;
        case 'T': /* Begin Trajectory control on all pucks */
            fer(cnt, npucks)
            {
                SCstarttrj(&(s[motor_position[cnt]]), commands[cnt]);
            }
            break;
        case 'c': /* Enter trajectory end position for present puck */
            start_entry();
            addstr("Enter next command: ");
            refresh();
            scanw("%lf", &(commands[cpuck]));
            finish_entry();
            break;
        case 'N': /* Enter gravity compensation constants by hand */
            start_entry();
            addstr("Enter GComp constants \"A B C D\": ");
            refresh();
            ret = scanw("%lf %lf %lf %lf", &A,&B,&C,&D);
            
            setGcompValues(A,B,C,D);
            finish_entry();
            break;

        case 'C': /* Enter trajectory end position for all pucks */
            start_entry();
            addstr("Enter next command for all pucks: ");
            refresh();
            scanw("%lf", &ftmp);
            fer (cnt, npucks)
            {
                commands[cnt] = ftmp;
            }
            finish_entry();
            break;
        case '.':   /* Play through playlist once */
            MLplay(&ml);
            break;
        case ',': /* Execute present line in playlist */
            MLone(&ml);
            break;
        case '>': /* Show next line in playlist */
            MLnext(&ml);
            break;
        case '<': /* Show previous line in playlist */
            MLprev(&ml);
            break;
        case '?':   /* Loop through playlist */
            MLrepeatplay(&ml);
            break;
        case '/':   /* Stop playlist execution */
            MLstop(&ml);
            break;
        case 'f':   /* Load playlist from file */
            start_entry();
            addstr("Enter filename for playlist read: ");
            refresh();
            scanw("%s", fn);
            MLload(&ml, fn);
            finish_entry();
            break;
        case 'F':   /* Save current playlist to file */
            start_entry();
            addstr("Enter filename for playlist save: ");
            refresh();
            scanw("%s", fn);
            MLsave(&ml, fn);
            finish_entry();
            break;
        case '+':   /* Add to playlist */
            MLaddSC(&ml);
            break;
        case '-':   /* Delete from playlist */
            MLdel(&ml);
            break;
/*        case 'h':   // Toggle help display 
            clear();
            disp_help = !disp_help;
            if (disp_help) present_screen = HELP_SCREEN;
            break;*/
/*        case 'g':
            clear();
            if (present_screen == MAIN_SCREEN)
              present_screen = GCOMP_SCREEN;
            else if (present_screen == GCOMP_SCREEN)
              present_screen = MAIN_SCREEN;
          
          break;*/
        case 'G':   /* Get temperatures */
            fer(cnt, npucks)
            {
                temperatures[act[cnt].puck.ID] = -1;
                tmp = GetProp(cnt, MOTOR_TEMPERATURE);
                //syslog(LOG_ERR,"%d Val1 %d",cnt,tmp);
                temperatures[act[cnt].puck.ID] = tmp;
                //syslog(LOG_ERR,"Val2 %d",modes[act[cnt].puck.ID]);
            }
            break;
        case 'E':   /* Get Saved bus error */
            getProperty(&(bus[act[0].bus].CANdev), SAFETY_MODULE, SAVED_BUS_ERR, &saved_bus_error);
            setProperty(&(bus[act[0].bus].CANdev), SAFETY_MODULE, SAVED_BUS_ERR, 0);
            refresh();
            break;
/*        case 'j':
          fer(cnt, npucks)
          {
            offsets[cnt] = GetProp(cnt,POSITION_OFFSET);
          }
          break;*/
        case 'H': /* Home / ParkWAM */
            ParkWAM();
            break;
        case 27: //Handle and discard extended keyboard characters (like arrows)
            if ((chr = getch()) != ERR)
            {
               if (chr == 91)
               {
                 if ((chr = getch()) != ERR) 
                 {
                    if (chr == 67) //Right arrow
                    {
                       cpuck++;
                       if (cpuck >= npucks)
                         cpuck = 0;
                    }
                    else if (chr == 68) //Left arrow
                    {
                      cpuck--;
                      if (cpuck < 0)
                        cpuck = npucks - 1;
                    }
                    else
                      while(getch()!=ERR){}
                 }
               }
               else
                while(getch()!=ERR){}
              }
              break;
        default:
           while(getch()!=ERR){}
           //syslog(LOG_ERR,"Caught unknown keyhit %d",c);
            break;
        //
    }
} //}}}


/** Evaluate a playlist
  
  checks to see if we have reached the next point and loads a new trajectory if we have. 
*/
void WACKYeval()
{
  int ret,cnt;
  double pos,vel,acc;
  
  if (WACKYplay)
  {
    ret = 0;
    for (cnt = 0; cnt < npucks; cnt++)
    {
      if (s[cnt].trj.state == 0){
        pos = (0.5-drand48()) * 1.4 * 2;
        //vel = (0.5-drand48()) * 1.4 * 2;
        SCstarttrj(&(s[cnt]), pos);
      }
    }
  }
}
/*======================================================================*
 *                                                                      *
 *             Copyright (c) 2003 Barrett Technology, Inc.              *
 *                        139 Main Street                               *
 *                       Kendall/MIT Square                             *
 *                  Cambridge, MA  02142-1528  USA                      *
 *                                                                      *
 *                        All rights reserved.                          *
 *                                                                      *
 *  ******************************************************************  *
 *                            DISCLAIMER                                *
 *                                                                      *
 *  This software and related documentation are provided to you on      *
 *  an as is basis and without warranty of any kind.  No warranties,    *
 *  express or implied, including, without limitation, any warranties   *
 *  of merchantability or fitness for a particular purpose are being    *
 *  provided by Barrett Technology, Inc.  In no event shall Barrett     *
 *  Technology, Inc. be liable for any lost development expenses, lost  *
 *  lost profits, or any incidental, special, or consequential damage.  *
 *                                                                      *
 *                        SUPPORTED COMPILERS                           *
 *                                                                      *
 *  This software has been compiled and tested with the following       *
 *  compilers:                                                          *
 *                                                                      
 *  - QNX qcc          
 *                                                                      *
 *  ******************************************************************  *
 *                                                                      
 * CVS info: $Id$
 * CVS automatic log (prune as desired):
 * $Log$
 *
 *======================================================================*/
